using System;
using System.Collections.Generic;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using LighthouseShared;
using System.Text.RegularExpressions;
using System.IO;
using System.Xml;
using System.Reflection;
using System.Runtime.InteropServices;

namespace Lighthouse
{
    public partial class FormMain : Form, IPluginHost
    {
        public string m_CWD;

        public string m_ProfilePath;
        public FileStream m_Settings;

        public string m_Chat_Name;
        public ushort m_Chat_Exchange;
        public ushort m_Chat_Instance;

        public string m_Client_Owner_Filename;
        public IClientPlugin m_Client_Owner;

        public string m_Client_CCOM_Filename;
        public IClientPlugin m_Client_CCOM;

        public string m_CCOM_Filename;
        public ICCOMPlugin m_CCOM;

        public List<string> m_GeneralPlugins_Filenames;
        public List<string> m_GeneralPlugins_Filenames_Loaded;
        public List<IGeneralPlugin> m_GeneralPlugins;

        public Dictionary<IPlugin, List<LHPanel>> m_UIPanels;

        public IChatInfo m_ChatInfo_Owner;
        public IChatInfo m_ChatInfo_CCOM;

        public List<CallbackContainer> m_Callbacks_ChatMessage;
        public List<CallbackContainer> m_Callbacks_ChatUserJoin;
        public List<CallbackContainer> m_Callbacks_ChatUserLeave;

        public List<MessageItem> m_ChatMessageQueue;
        public int m_ChatMessageLastTime;
        public bool m_ChatMessageWaiting;

        public Queue<string> m_BanQueue;
        public int m_BanLastTime;
        public Queue<string> m_KickQueue;
        public int m_KickLastTime;
        public Queue<string> m_UnbanQueue;
        public int m_UnbanLastTime;

        public int m_ChatMessageMax = 20;

        public Control buttonScrollUp;
        public Control buttonScrollDown;
        public bool m_Scrolling;

        public System.Threading.Timer m_TimeoutTimer;

        public FormMain(string profilepath, FileStream settings)
        {
            InitializeComponent();

            this.MouseWheel += new MouseEventHandler(FormMain_MouseWheel);

            panelButtonBar.MouseEnter += new EventHandler(panelButtonBar_MouseEnter);

            buttonScrollUp = new Control();
            buttonScrollUp.Width = panelButtonBar.Width / 2;
            buttonScrollUp.Height = 14;
            buttonScrollUp.BackColor = SystemColors.ControlDark;
            buttonScrollUp.Top = (panelButtonBar.Top + panelButtonBar.Height) - buttonScrollUp.Height;
            buttonScrollUp.Left = panelButtonBar.Left;
            buttonScrollUp.MouseUp += new MouseEventHandler(buttonScrollUp_MouseUp);
            buttonScrollUp.MouseDown += new MouseEventHandler(buttonScrollUp_MouseDown);
            buttonScrollUp.Paint += new PaintEventHandler(buttonScroll_Paint);

            buttonScrollDown = new Control();
            buttonScrollDown.Width = panelButtonBar.Width / 2;
            buttonScrollDown.Height = 14;
            buttonScrollDown.BackColor = SystemColors.ControlDark;
            buttonScrollDown.Top = (panelButtonBar.Top + panelButtonBar.Height) - buttonScrollDown.Height;
            buttonScrollDown.Left = panelButtonBar.Left + buttonScrollDown.Width;
            buttonScrollDown.MouseDown += new MouseEventHandler(buttonScrollDown_MouseDown);
            buttonScrollDown.MouseUp += new MouseEventHandler(buttonScrollDown_MouseUp);
            buttonScrollDown.Paint += new PaintEventHandler(buttonScroll_Paint);


            buttonShowSignOn.Tag = lhPanelSignOn;
            buttonShowPlugins.Tag = lhPanelPlugins;
            buttonShowChat.Tag = lhPanelChat;
            buttonShowLog.Tag = lhPanelLog;

            m_CWD = Directory.GetCurrentDirectory();

            m_ProfilePath = profilepath;
            m_Settings = settings;

            m_GeneralPlugins_Filenames_Loaded = new List<string>();
            m_GeneralPlugins_Filenames = new List<string>();
            m_GeneralPlugins = new List<IGeneralPlugin>();

            m_UIPanels = new Dictionary<IPlugin, List<LHPanel>>();

            m_Callbacks_ChatMessage = new List<CallbackContainer>();
            m_Callbacks_ChatUserJoin = new List<CallbackContainer>();
            m_Callbacks_ChatUserLeave = new List<CallbackContainer>();

            m_ChatMessageQueue = new List<MessageItem>();

            m_BanQueue = new Queue<string>();
            m_KickQueue = new Queue<string>();
            m_UnbanQueue = new Queue<string>();

            m_TimeoutTimer = new System.Threading.Timer(new System.Threading.TimerCallback(TimeoutTimer_Callback));
            m_TimeoutTimer.Change(3000, 3000);

            //default values for UI things
            labelAppTitle.Text = "Project Lighthouse 2.0 Alpha 2";
            comboLogLevel.SelectedIndex = (int)LogLevel.Normal;
            checkBoxSaveStatusLog.Checked = true;


            ReadSettings();
            InitPlugins();

            UpdateUI();
        }

        void TimeoutTimer_Callback(object state)
        {

            try
            {
                if (Environment.TickCount - m_BanLastTime > 6000)
                {
                    m_BanLastTime = Environment.TickCount;

                    if (m_BanQueue.Count > 0)
                        m_BanQueue.Dequeue();

                    if (m_BanQueue.Count > 0)
                        m_Client_Owner.ChatBan(m_ChatInfo_Owner, m_BanQueue.Peek());
                }

                if (Environment.TickCount - m_KickLastTime > 6000)
                {
                    m_KickLastTime = Environment.TickCount;

                    if (m_KickQueue.Count > 0)
                        m_KickQueue.Dequeue();

                    if (m_KickQueue.Count > 0)
                        m_Client_Owner.ChatKick(m_ChatInfo_Owner, m_KickQueue.Peek());
                }

                if (Environment.TickCount - m_UnbanLastTime > 6000)
                {
                    m_UnbanLastTime = Environment.TickCount;

                    if (m_UnbanQueue.Count > 0)
                        m_UnbanQueue.Dequeue();

                    if (m_UnbanQueue.Count > 0)
                        m_Client_Owner.ChatUnban(m_ChatInfo_Owner, m_UnbanQueue.Peek());
                }

                if (Environment.TickCount - m_ChatMessageLastTime > 6000)
                {
                    lock (m_ChatMessageQueue)
                    {
                        if (!m_ChatMessageWaiting && m_ChatMessageQueue.Count > 0)
                        {
                            m_ChatMessageLastTime = Environment.TickCount;
                            m_Client_CCOM.ChatMessage(m_ChatInfo_CCOM, m_ChatMessageQueue[0].Data);
                            m_ChatMessageQueue.RemoveAt(0);
                        }
                    }
                }
            }
            catch (Exception)
            {
            }
        }


        void buttonScrollDown_MouseUp(object sender, MouseEventArgs e)
        {
            m_Scrolling = false;
        }

        void buttonScrollDown_MouseDown(object sender, MouseEventArgs e)
        {
            m_Scrolling = true;

            while (m_Scrolling && panelButtonBar.Controls[panelButtonBar.Controls.Count - 1].Top + 22 > panelButtonBar.Height + 10)
            {
                foreach (Control c in panelButtonBar.Controls)
                {
                    c.Top -= 22;
                }
                Application.DoEvents();
                System.Threading.Thread.Sleep(250);
                Application.DoEvents();
            }
        }

        void buttonScrollUp_MouseUp(object sender, MouseEventArgs e)
        {
            m_Scrolling = false;
        }

        void buttonScrollUp_MouseDown(object sender, MouseEventArgs e)
        {
            m_Scrolling = true;

            while (m_Scrolling && panelButtonBar.Controls[0].Top < -22)
            {
                foreach (Control c in panelButtonBar.Controls)
                {
                    c.Top += 22;
                }
                Application.DoEvents();
                System.Threading.Thread.Sleep(250);
                Application.DoEvents();
            }
        }

        void buttonScroll_Paint(object sender, PaintEventArgs e)
        {
            //get all the information from the control for drawing
            Control control = (Control)sender;
            Graphics g = e.Graphics;
            g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
            Pen p = new Pen(Color.Black, 1);
            p.StartCap = System.Drawing.Drawing2D.LineCap.Flat;
            p.EndCap = System.Drawing.Drawing2D.LineCap.Flat;

            // draw the arrow
            if (buttonScrollDown == sender)
            {
                g.DrawLine(p, 22, 4, 27, 10);
                g.DrawLine(p, 27, 10, 32, 4);
            }
            else if (buttonScrollUp == sender)
            {
                g.DrawLine(p, 32, 10, 27, 4);
                g.DrawLine(p, 27, 4, 22, 10);
            }
            
            //draw the borders
            g.DrawLine(p, 0, control.Height - 1, control.Width - 1, control.Height - 1);
            g.DrawLine(p, 0, 0, 0, control.Height - 1);
            //if its the button on the right, draw a border on the right
            if (sender.Equals(buttonScrollDown))
                g.DrawLine(p, control.Width - 1, 0, control.Width - 1, control.Height - 1);

            //clean up
            p.Dispose();
        }

        void panelButtonBar_MouseEnter(object sender, EventArgs e)
        {
            panelButtonBar.Focus();
        }

        void FormMain_MouseWheel(object sender, MouseEventArgs e)
        {
            if ((e.X > panelButtonBar.Left && e.X < (panelButtonBar.Left + panelButtonBar.Width)) && (e.Y > panelButtonBar.Top && e.Y < (panelButtonBar.Top + panelButtonBar.Height)))
            {
                if (e.Delta > 0 && panelButtonBar.Controls[0].Top < -22)
                {
                    foreach (Control c in panelButtonBar.Controls)
                    {
                        c.Top += 22;
                    }
                }
                else if (e.Delta < 0 && panelButtonBar.Controls[panelButtonBar.Controls.Count - 1].Top + 22 > panelButtonBar.Height + 10)
                {
                    foreach (Control c in panelButtonBar.Controls)
                    {
                        c.Top -= 22;
                    }
                }
            }
                 
        }


        public void Log(LogLevel loglevel, string message)
        {
            this.Invoke(new MethodInvoker(delegate
            {
                string datetime = DateTime.Now.ToString();

                if (checkBoxSaveStatusLog.Checked)
                {
                    File.AppendAllText(m_ProfilePath + @"\log.txt", String.Format("{0}|{1}|{2}\r\n", (int)loglevel, datetime, message));
                }
                if (comboLogLevel.SelectedIndex >= (int)loglevel)
                {
                    listLog.Items.Add(String.Format("[{0}] {1}", datetime, message));

                    if (listLog.SelectedIndex == listLog.Items.Count)
                        listLog.SelectedIndex++;
                }
            }));
        }

        public void ReadSettings()
        {
            XmlDocument xml = new XmlDocument();

            StreamReader sr = new StreamReader(m_Settings);
            xml.LoadXml(sr.ReadToEnd());

            foreach (XmlElement x in xml["settings"])
            {
                switch (x.Name.ToLower())
                {
                    case "chat":
                        foreach (XmlElement chatx in x)
                        {
                            switch (chatx.Name.ToLower())
                            {
                                case "name":
                                    m_Chat_Name = chatx.InnerText;
                                    break;
                                case "exchange":
                                    m_Chat_Exchange = ushort.Parse(chatx.InnerText);
                                    break;
                                case "instance":
                                    m_Chat_Instance = ushort.Parse(chatx.InnerText);
                                    break;
                            }
                        }
                        break;

                    case "sysplugins":
                        foreach (XmlElement syspluginsx in x)
                        {
                            switch (syspluginsx.Name.ToLower())
                            {
                                case "ownerclient":
                                    m_Client_Owner_Filename = syspluginsx.InnerText;
                                    break;
                                case "ccomclient":
                                    m_Client_CCOM_Filename = syspluginsx.InnerText;
                                    break;
                                case "commandparser":
                                    m_CCOM_Filename = syspluginsx.InnerText;
                                    break;
                            }
                        }
                        break;

                    case "genplugins":
                        m_GeneralPlugins_Filenames_Loaded.Clear();
                        foreach (XmlElement genpluginsx in x)
                        {
                            if (genpluginsx.Name.ToLower() == "plugin")
                            {
                                m_GeneralPlugins_Filenames.Add(genpluginsx.InnerText);
                                if (genpluginsx.Attributes["load"].Value.ToLower() == "true")
                                {
                                    m_GeneralPlugins_Filenames_Loaded.Add(genpluginsx.InnerText);
                                }
                            }
                        }
                        break;
                }
            }
        }

        public void SaveSettings()
        {
            
        }

        public void InitPlugins()
        {
            m_Client_Owner = (IClientPlugin)LoadPluginFromFile(m_CWD + @"\Clients\" + m_Client_Owner_Filename, typeof(IClientPlugin), m_ProfilePath + @"\ClientOwner.{0}-{1}.settings.xml");
            m_Client_Owner.OnSignedOn += new SignOnHandler(m_Client_Owner_OnSignedOn);
            m_Client_Owner.OnSignedOff += new SignOffHandler(m_Client_Owner_OnSignedOff);
            m_Client_Owner.OnStatusUpdate += new StatusUpdateHandler(m_Client_Owner_OnStatusUpdate);
            m_Client_Owner.OnChatJoin += new ChatJoinHandler(m_Client_Owner_OnChatJoin);
            m_Client_Owner.OnChatLeave += new ChatLeaveHandler(m_Client_Owner_OnChatLeave);
            m_Client_Owner.OnChatBan += new ChatUserHandler(m_Client_Owner_OnChatBan);
            m_Client_Owner.OnChatUnbanFailed += new ChatUserHandler(m_Client_Owner_OnChatUnbanFailed);
            m_Client_Owner.OnChatKick += new ChatUserHandler(m_Client_Owner_OnChatKick);
            m_Client_Owner.OnChatKickFailed += new ChatUserHandler(m_Client_Owner_OnChatKickFailed);
            m_Client_Owner.OnChatBanFailed += new ChatUserHandler(m_Client_Owner_OnChatBanFailed);
            m_Client_Owner.OnChatUnban += new ChatUserHandler(m_Client_Owner_OnChatUnban);
            m_Client_Owner.OnChatMessage += new ChatMessageHandler(m_Client_Owner_OnChatMessage);
            m_Client_CCOM = (IClientPlugin)LoadPluginFromFile(m_CWD + @"\Clients\" + m_Client_CCOM_Filename, typeof(IClientPlugin), m_ProfilePath + @"\ClientCCOM.{0}-{1}.settings.xml");
            m_Client_CCOM.OnSignedOn += new SignOnHandler(m_Client_CCOM_OnSignedOn);
            m_Client_CCOM.OnSignedOff += new SignOffHandler(m_Client_CCOM_OnSignedOff);
            m_Client_CCOM.OnStatusUpdate += new StatusUpdateHandler(m_Client_CCOM_OnStatusUpdate);
            m_Client_CCOM.OnChatJoin += new ChatJoinHandler(m_Client_CCOM_OnChatJoin);
            m_Client_CCOM.OnChatUserJoin += new ChatUserHandler(m_Client_CCOM_OnChatUserJoin);
            m_Client_CCOM.OnChatUserLeave += new ChatUserHandler(m_Client_CCOM_OnChatUserLeave);
            m_Client_CCOM.OnChatOwnerDetect += new ChatUserHandler(m_Client_CCOM_OnChatOwnerDetect);
            m_Client_CCOM.OnChatMessage += new ChatMessageHandler(m_Client_CCOM_OnChatMessage);
            m_Client_CCOM.OnChatLeave += new ChatLeaveHandler(m_Client_CCOM_OnChatLeave);
            m_CCOM = (ICCOMPlugin)LoadPluginFromFile(m_CWD + @"\CommandParsers\" + m_CCOM_Filename, typeof(ICCOMPlugin), m_ProfilePath + @"\CCOMPlugin.{0}-{1}.settings.xml");

            LoadGeneralPlugins();

            UpdateUI();
        }

        public void LoadGeneralPlugins()
        {
            bool errors = false;

            foreach (string s in m_GeneralPlugins_Filenames_Loaded)
            {
                try
                {
                    m_GeneralPlugins.Add((IGeneralPlugin)LoadPluginFromFile(m_CWD + @"\Plugins\" + s, typeof(IGeneralPlugin), m_ProfilePath + @"\Plugin.{0}-{1}.settings.xml"));
                }
                catch (Exception)
                {
                    errors = true;
                }
            }

            if (errors)
            {
                //*** do something to warn the user about errors
            }

            UpdateUI();
        }

        public void UnloadGeneralPlugins()
        {
            bool errors = false;

            foreach (IPlugin plugin in m_GeneralPlugins)
            {
                CallbackContainer[] tmpccs;

                tmpccs = m_Callbacks_ChatMessage.ToArray();
                
                foreach (CallbackContainer cc in tmpccs)
                {
                    if (cc.Plugin == plugin)
                        m_Callbacks_ChatMessage.Remove(cc);
                }

                tmpccs = m_Callbacks_ChatUserJoin.ToArray();

                foreach (CallbackContainer cc in tmpccs)
                {
                    if (cc.Plugin == plugin)
                        m_Callbacks_ChatUserJoin.Remove(cc);
                }

                tmpccs = m_Callbacks_ChatUserLeave.ToArray();

                foreach (CallbackContainer cc in tmpccs)
                {
                    if (cc.Plugin == plugin)
                        m_Callbacks_ChatUserLeave.Remove(cc);
                }

                m_CCOM.UnregisterCommandsByPlugin(plugin);

                RemoveUIPanelsByPlugin(plugin);

                try
                {
                    plugin.SaveSettings();
                    plugin.Unload();
                }
                catch (Exception)
                {
                    errors = true;
                }
            }
            m_GeneralPlugins.Clear();

            if (errors)
            {
                //*** do something to warn the user about errors
            }

            UpdateUI();
        }

        void m_Client_Owner_OnChatMessage(IClientPlugin sender, IChatInfo chatinfo, string screenname, ulong icbmcookie, string message)
        {

        }

        void m_Client_Owner_OnChatKickFailed(IClientPlugin sender, IChatInfo chatinfo, string screenname)
        {
            if (m_KickQueue.Count > 0)
                m_KickQueue.Dequeue();

            if (m_KickQueue.Count > 0)
                m_Client_Owner.ChatKick(m_ChatInfo_Owner, m_KickQueue.Peek());

            CCOMPlugin.HandleKickFailed(screenname);
        }

        void m_Client_Owner_OnChatLeave(IClientPlugin sender, IChatInfo chatinfo, bool banned, bool closed)
        {
            this.Invoke(new MethodInvoker(delegate
            {
                try
                {
                    buttonJoinOwner.Text = "Join Owner";
                    labelOwnerStatus.Text = "Left " + chatinfo.DisplayName;
                    Log(LogLevel.Minimal, "Owner Left " + chatinfo.DisplayName);
                }
                catch (Exception)
                {
                }
            }));

            if (checkOwnerAutoRejoin.Checked)
                m_Client_Owner.ChatJoin(textRoomName.Text, ushort.Parse(textExchange.Text));
        }

        void m_Client_Owner_OnChatBanFailed(IClientPlugin sender, IChatInfo chatinfo, string screenname)
        {
            if (m_BanQueue.Count > 0)
                m_BanQueue.Dequeue();

            if (m_BanQueue.Count > 0)
                m_Client_Owner.ChatBan(m_ChatInfo_Owner, m_BanQueue.Peek());

            CCOMPlugin.HandleBanFailed(screenname);
        }

        void m_Client_Owner_OnChatUnbanFailed(IClientPlugin sender, IChatInfo chatinfo, string screenname)
        {
            if (m_UnbanQueue.Count > 0)
                m_UnbanQueue.Dequeue();

            if (m_UnbanQueue.Count > 0)
                m_Client_Owner.ChatUnban(m_ChatInfo_Owner, m_UnbanQueue.Peek());

            CCOMPlugin.HandleUnbanFailed(screenname);
        }

        void m_Client_Owner_OnChatUnban(IClientPlugin sender, IChatInfo chatinfo, string screenname)
        {
            if (m_UnbanQueue.Count > 0)
                m_UnbanQueue.Dequeue();

            if (m_UnbanQueue.Count > 0)
                m_Client_Owner.ChatUnban(m_ChatInfo_Owner, m_UnbanQueue.Peek());

            CCOMPlugin.HandleUnban(screenname);
        }

        void m_Client_Owner_OnChatKick(IClientPlugin sender, IChatInfo chatinfo, string screenname)
        {
            if (m_KickQueue.Count > 0)
                m_KickQueue.Dequeue();

            if (m_KickQueue.Count > 0)
                m_Client_Owner.ChatKick(m_ChatInfo_Owner, m_KickQueue.Peek());

            CCOMPlugin.HandleKick(screenname);
        }

        void m_Client_Owner_OnChatBan(IClientPlugin sender, IChatInfo chatinfo, string screenname)
        {
            if (m_BanQueue.Count > 0)
                m_BanQueue.Dequeue();

            if (m_BanQueue.Count > 0)
                m_Client_Owner.ChatBan(m_ChatInfo_Owner, m_BanQueue.Peek());

            CCOMPlugin.HandleBan(screenname);
        }

        void m_Client_CCOM_OnChatLeave(IClientPlugin sender, IChatInfo chatinfo, bool banned, bool closed)
        {
            this.Invoke(new MethodInvoker(delegate
            {
                try
                {
                    buttonJoinCCOM.Text = "Join CCOM";
                    labelCCOMStatus.Text = "Left " + chatinfo.DisplayName;
                    Log(LogLevel.Minimal, "CCOM Left " + chatinfo.DisplayName);
                }
                catch (Exception)
                {
                }
            }));

            if (checkCCOMAutoRejoin.Checked)
                m_Client_CCOM.ChatJoin(textRoomName.Text, ushort.Parse(textExchange.Text));

        }

        void m_Client_CCOM_OnChatMessage(IClientPlugin sender, IChatInfo chatinfo, string screenname, ulong icbmcookie, string message)
        {
            foreach (CallbackContainer cc in m_Callbacks_ChatMessage)
            {
                if (((ChatMessageCallback)cc.Callback)(chatinfo, ref screenname, ref message))
                    break;
            }


            lock (m_ChatMessageQueue)
            {
                if (m_ChatMessageQueue.Count > 0 && screenname == m_Client_CCOM.ScreenName)
                {
                    if (!m_ChatMessageWaiting)
                    {
                        if (Environment.TickCount - m_ChatMessageLastTime < 800)
                        {
                            m_ChatMessageWaiting = true;
                            System.Threading.ThreadPool.QueueUserWorkItem(new System.Threading.WaitCallback(delegate(object o)
                            {
                                try
                                {
                                    System.Threading.Thread.Sleep(800);
                                    lock (m_ChatMessageQueue)
                                    {
                                        m_ChatMessageWaiting = false;
                                        m_ChatMessageLastTime = Environment.TickCount;
                                        m_Client_CCOM.ChatMessage(m_ChatInfo_CCOM, m_ChatMessageQueue[0].Data);
                                        m_ChatMessageQueue.RemoveAt(0);
                                    }
                                }
                                catch (Exception)
                                {
                                }
                            }));
                        }
                        else
                        {
                            m_ChatMessageLastTime = Environment.TickCount;
                            m_Client_CCOM.ChatMessage(m_ChatInfo_CCOM, m_ChatMessageQueue[0].Data);
                            m_ChatMessageQueue.RemoveAt(0);
                        }
                    }
                }
            }

            m_CCOM.HandleMessage(screenname, message);
        }

        void m_Client_CCOM_OnChatOwnerDetect(IClientPlugin sender, IChatInfo chatinfo, string screenname)
        {

        }

        void m_Client_CCOM_OnChatUserLeave(IClientPlugin sender, IChatInfo chatinfo, string screenname)
        {
            if (checkWatchOwner.Checked && m_Client_Owner.ScreenName == screenname.ToLower().Replace(" ", ""))
            {
                m_Client_Owner.ChatJoin(textRoomName.Text, ushort.Parse(textExchange.Text));
            }

            foreach (CallbackContainer cc in m_Callbacks_ChatUserLeave)
            {
                if (((ChatUserCallback)cc.Callback)(chatinfo, ref screenname))
                    break;
            }

            m_CCOM.HandleUserLeave(screenname.ToLower().Replace(" ", ""));
        }

        void m_Client_CCOM_OnChatUserJoin(IClientPlugin sender, IChatInfo chatinfo, string screenname)
        {
            m_CCOM.HandleUserJoin(screenname);

            foreach (CallbackContainer cc in m_Callbacks_ChatUserJoin)
            {
                if (((ChatUserCallback)cc.Callback)(chatinfo, ref screenname))
                    break;
            }   
        }

        void m_Client_CCOM_OnChatJoin(IClientPlugin sender, IChatInfo chatinfo)
        {

            m_ChatInfo_CCOM = chatinfo;

            this.Invoke(new MethodInvoker(delegate
            {
                buttonJoinCCOM.Text = "Leave CCOM";
                labelCCOMStatus.Text = "Joined " + chatinfo.DisplayName;
                Log(LogLevel.Minimal, "CCOM Joined " + chatinfo.DisplayName);
                lock (m_ChatMessageQueue)
                {
                    m_ChatMessageQueue.Clear();
                }
            }));
        }

        void m_Client_CCOM_OnStatusUpdate(IClientPlugin sender, string status)
        {
            this.Invoke(new MethodInvoker(delegate
            {
                labelCCOMClientStatus.Text = status;
            }));
        }

        void m_Client_CCOM_OnSignedOff(IClientPlugin sender)
        {
            this.Invoke(new MethodInvoker(delegate
            {
                buttonCCOMSignOn.Text = "Sign On";
            }));
        }

        void m_Client_CCOM_OnSignedOn(IClientPlugin sender)
        {
            this.Invoke(new MethodInvoker(delegate
            {
                buttonCCOMSignOn.Text = "Sign Off";

                if (checkCCOMJoinSignOn.CheckState == CheckState.Checked)
                {
                    m_Client_CCOM.ChatJoin(textRoomName.Text, ushort.Parse(textExchange.Text));
                }
            }));
        }

        void m_Client_Owner_OnChatOwnerDetect(IClientPlugin sender, IChatInfo chatinfo, string screenname)
        {
            
        }

        void m_Client_Owner_OnChatJoin(IClientPlugin sender, IChatInfo chatinfo)
        {
            m_ChatInfo_Owner = chatinfo;
            this.Invoke(new MethodInvoker(delegate
            {
                buttonJoinOwner.Text = "Leave Owner";
                labelOwnerStatus.Text = "Joined " + chatinfo.DisplayName;
                Log(LogLevel.Minimal, "Owner Joined " + chatinfo.DisplayName);
            }));
        }

        void m_Client_Owner_OnStatusUpdate(IClientPlugin sender, string status)
        {
            this.Invoke(new MethodInvoker(delegate
            {
                labelOwnerClientStatus.Text = status;
            }));
        }

        void m_Client_Owner_OnSignedOff(IClientPlugin sender)
        {
            this.Invoke(new MethodInvoker(delegate
            {
                buttonOwnerSignOn.Text = "Sign On";
            }));
        }

        void m_Client_Owner_OnSignedOn(IClientPlugin sender)
        {
            this.Invoke(new MethodInvoker(delegate
            {
                buttonOwnerSignOn.Text = "Sign Off";

                if (checkOwnerJoinSignOn.CheckState == CheckState.Checked)
                {
                    m_Client_Owner.ChatJoin(textRoomName.Text, ushort.Parse(textExchange.Text));
                }
            }));

            this.Log(LogLevel.Minimal, "Signed On Owner: " + sender.ScreenName);
        }

        private object LoadPluginFromFile(string path, Type plugintype, string formatfilename)
        {
            Assembly asm = Assembly.LoadFile(path);

            foreach (Type t in asm.GetTypes())
            {
                if (t.GetInterface(plugintype.Name.ToString()) != null)
                {
                    // do some weird shit that i dont understand, but microsoft says it so it must be true
                    object plugin = Activator.CreateInstance(t);
                    ((IPlugin)plugin).Load(this, new FileStream(String.Format(formatfilename, ((IPlugin)plugin).Name.Replace(" ", ""), ((IPlugin)plugin).Version), FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.Read));
                    return plugin;
                }
            }

            return null;
        }

        public void UpdateUI()
        {
            try
            {
                textOwnerClientPlugin.Text = ((IPlugin)m_Client_Owner).Name;
                textCCOMClientPlugin.Text = ((IPlugin)m_Client_CCOM).Name;
                textCCOMPlugin.Text = ((IPlugin)m_CCOM).Name;

                listBoxGeneralPlugins.Items.Clear();
                foreach (IGeneralPlugin p in m_GeneralPlugins)
                {
                    listBoxGeneralPlugins.Items.Add(((IPlugin)p).Name);
                }

                int buttonbarheight = 0;
                int buttontop = -1; // unbutton my top! lolz
                foreach (Control c in panelButtonBar.Controls)
                {
                    c.Top = buttontop;
                    buttontop += 22;
                    buttonbarheight += c.Height;
                }
                panelButtonBar.Height = Math.Min(panelButtonBar.MaximumSize.Height, buttonbarheight) - (panelButtonBar.Controls.Count - 1);

                if (buttontop > panelButtonBar.MaximumSize.Height)
                {
                    panelButtonBar.Height = panelButtonBar.MaximumSize.Height - 14;

                    this.Controls.Add(buttonScrollUp);
                    this.Controls.Add(buttonScrollDown);

                }
                else
                {
                    panelButtonBar.Top = 8;
                }
            }
            catch (Exception)
            {
            }
        }

        private void buttonShowPanel_Click(object sender, EventArgs e)
        {
            Button buttonclicked = (Button)sender;
            LHPanel showpanel = (LHPanel)buttonclicked.Tag;

            foreach (Control c in panelButtonBar.Controls)
            {
                c.BackColor = Color.LightSteelBlue;
            }
            buttonclicked.BackColor = Color.SteelBlue;

            showpanel.UpdateUI();
            showpanel.BringToFront();
            showpanel.Focus();
        }

        private void buttonCCOMSignOn_Click(object sender, EventArgs e)
        {
            if (buttonCCOMSignOn.Text == "Sign On")
            {
                m_Client_CCOM.SignOn(textCCOMScreenName.Text, textCCOMPassword.Text);
            }
            else if (buttonCCOMSignOn.Text == "Sign Off")
            {
                m_Client_CCOM.SignOff();
            }
        }

        private void buttonJoinCCOM_Click(object sender, EventArgs e)
        {
            if (buttonJoinCCOM.Text == "Join CCOM")
            {
                m_Client_CCOM.ChatJoin(textRoomName.Text, ushort.Parse(textExchange.Text));
            }
            else if (buttonJoinCCOM.Text == "Leave CCOM")
            {
                m_Client_CCOM.ChatLeave(m_ChatInfo_CCOM);
            }
        }

        private void buttonOwnerSignOn_Click(object sender, EventArgs e)
        {
            if (buttonOwnerSignOn.Text == "Sign On")
            {
                m_Client_Owner.SignOn(textOwnerScreenName.Text, textOwnerPassword.Text);
            }
            else if (buttonOwnerSignOn.Text == "Sign Off")
            {
                m_Client_Owner.SignOff();
            }
        }

        private void buttonJoinOwner_Click(object sender, EventArgs e)
        {
            if (buttonJoinOwner.Text == "Join Owner")
            {
                try
                {
                    m_Client_Owner.ChatJoin(textRoomName.Text, ushort.Parse(textExchange.Text));
                }
                catch (Exception ex)
                {
                    MessageBox.Show(this, ex.Message, "Cannot Join Chat");
                }
            }
            else if (buttonJoinOwner.Text == "Leave Owner")
            {
                m_Client_Owner.ChatLeave(m_ChatInfo_Owner);
            }
        }


        #region IPluginHost Members

        public object RegisterChatMessageHook(IPlugin plugin, ChatMessageCallback callback)
        {
            CallbackContainer cc = new CallbackContainer();
            cc.Callback = callback;
            cc.Plugin = plugin;
            m_Callbacks_ChatMessage.Add(cc);
            return cc;
        }

        public void UnregisterChatMessageHook(object handle)
        {
            m_Callbacks_ChatMessage.Remove((CallbackContainer)handle);
        }

        public object RegisterChatUserJoinHook(IPlugin plugin, ChatUserCallback callback)
        {
            CallbackContainer cc = new CallbackContainer();
            cc.Callback = callback;
            cc.Plugin = plugin;
            m_Callbacks_ChatUserJoin.Add(cc);
            return cc;
        }

        public void UnregisterChatUserJoinHook(object handle)
        {
            m_Callbacks_ChatUserJoin.Remove((CallbackContainer)handle);
        }

        public object RegisterChatUserLeaveHook(IPlugin plugin, ChatUserCallback callback)
        {
            CallbackContainer cc = new CallbackContainer();
            cc.Callback = callback;
            cc.Plugin = plugin;
            m_Callbacks_ChatUserLeave.Add(cc);
            return callback;
        }

        public void UnregisterChatUserLeaveHook(object handle)
        {
            m_Callbacks_ChatUserLeave.Remove((CallbackContainer)handle);
        }

        public void SendChatMessageByType(string type, string[] items, bool single)
        {
            try
            {
                lock (m_ChatMessageQueue)
                {
                    foreach (string item in items)
                    {
                        bool found = false;
                        foreach (MessageItem msgitem in m_ChatMessageQueue)
                        {
                            if (!msgitem.Single && msgitem.Type == type)
                            {
                                string newdata = msgitem.Data + ", " + item;

                                //*** this line needs to be updated to use MaxVisibleLength when i code it
                                if (newdata.Length <= 232)
                                {
                                    msgitem.Data = newdata;
                                    found = true;
                                }
                            }
                        }

                        if (found == false)
                        {
                            MessageItem msgitem = new MessageItem(type);

                            msgitem.Data = msgitem.Type + "  " + item;

                            m_ChatMessageQueue.Add(msgitem);
                        }
                    }
                }
                lock (m_ChatMessageQueue)
                {
                    if (!m_ChatMessageWaiting && m_ChatMessageQueue.Count == 1)
                    {
                        if (Environment.TickCount - m_ChatMessageLastTime < 800)
                        {
                            m_ChatMessageWaiting = true;
                            System.Threading.ThreadPool.QueueUserWorkItem(new System.Threading.WaitCallback(delegate(object o)
                            {
                                try
                                {
                                    System.Threading.Thread.Sleep(800);
                                    lock (m_ChatMessageQueue)
                                    {
                                        m_ChatMessageWaiting = false;
                                        m_ChatMessageLastTime = Environment.TickCount;
                                        m_Client_CCOM.ChatMessage(m_ChatInfo_CCOM, m_ChatMessageQueue[0].Data);
                                        m_ChatMessageQueue.RemoveAt(0);
                                    }
                                }
                                catch (Exception)
                                {
                                }
                            }));
                        }
                        else
                        {
                            m_ChatMessageLastTime = Environment.TickCount;
                            m_Client_CCOM.ChatMessage(m_ChatInfo_CCOM, m_ChatMessageQueue[0].Data);
                            m_ChatMessageQueue.RemoveAt(0);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                if (ex.GetType() == typeof(NullReferenceException))
                {
                    // **ERROR: cant send message, not in chat
                    lock (m_ChatMessageQueue)
                    {

                        m_ChatMessageQueue.Clear();
                    }
                }
                else
                {
                    // not sure why i had this being thrown, 
                    // but theres no reason for the whole deal to die if it cant send a message o_O\

                    //throw ex; // i threw my ex in the river i hope that bitch drowned
                }
            }

        }

        public void SendChatMessageByType(string type, string item, bool single)
        {
            SendChatMessageByType(type, new string[] { item }, single);            
        }

        public void SendChatMessageByType(string type, string item)
        {
            SendChatMessageByType(type, item, false);
        }

        public void SendChatMessageByType(string type, string[] items)
        {
            SendChatMessageByType(type, items, false);
        }

       


        #endregion


        #region IPluginHost Members


        public void SendBan(string screenname)
        {
            if (screenname == m_Client_CCOM.ScreenName)
                return;

            if (m_ChatInfo_Owner != null)
            {
                m_BanQueue.Enqueue(screenname);

                if (m_BanQueue.Count == 1)
                    m_Client_Owner.ChatBan(m_ChatInfo_Owner, screenname);
            }
            else
            {
                throw new OwnerNotInChatException();
            }
        }

        public void SendUnban(string screenname)
        {
            if (m_ChatInfo_Owner != null)
            {

                m_UnbanQueue.Enqueue(screenname);

                if (m_UnbanQueue.Count == 1)
                    m_Client_Owner.ChatUnban(m_ChatInfo_Owner, screenname);

            }
            else
            {
                throw new OwnerNotInChatException();
            }
        }

        public void SendKick(string screenname)
        {
            if (screenname == m_Client_CCOM.ScreenName)
                return;

            if (m_ChatInfo_Owner != null)
            {

                m_KickQueue.Enqueue(screenname);

                if (m_KickQueue.Count == 1)
                    m_Client_Owner.ChatKick(m_ChatInfo_Owner, screenname);

            }
            else
            {
                throw new OwnerNotInChatException();
            }
        }



        #endregion

        #region IPluginHost Members

        public void AddUIPanel(IPlugin plugin, LHPanel panel)
        {
            if (!m_UIPanels.ContainsKey(plugin))
                m_UIPanels.Add(plugin, new List<LHPanel>());

            m_UIPanels[plugin].Add(panel);

            this.Controls.Add(panel);
            panel.Left = panelBlank.Left;
            panel.Top = panelBlank.Top;

            Button showbutton = new Button();
            showbutton.Left = -1;
            showbutton.Width = 110;
            showbutton.Height = 23;
            showbutton.BackColor = Color.LightSteelBlue;
            showbutton.FlatStyle = FlatStyle.Flat;
            showbutton.Text = panel.PanelName;
            showbutton.Tag = panel;
            showbutton.Click += new EventHandler(buttonShowPanel_Click);

            panel.ShowButton = showbutton;

            panelButtonBar.Controls.Add(panel.ShowButton);

            UpdateUI();
        }

        public void RemoveUIPanel(IPlugin plugin, LHPanel panel)
        {
            this.Controls.Remove(panel);
            this.panelButtonBar.Controls.Remove(panel.ShowButton);

            m_UIPanels[plugin].Remove(panel);

            UpdateUI();
        }

        public void RemoveUIPanelsByPlugin(IPlugin plugin)
        {
            LHPanel[] tmppanels = m_UIPanels[plugin].ToArray();

            foreach (LHPanel panel in tmppanels)
            {
                RemoveUIPanel(plugin, panel);
            }

            m_UIPanels.Remove(plugin);
        }

        public IClientPlugin ClientCCOM
        {
            get { return m_Client_CCOM; }
        }

        public IClientPlugin ClientOwner
        {
            get { return m_Client_Owner; }
        }


        public ICommandInfo RegisterCommand(string name, string defaulttext, string help, IPlugin plugin, CommandCallback callback)
        {
            return m_CCOM.RegisterCommand(name, defaulttext, help, plugin, callback);
        }

        public void UnregisterCommand(ICommandInfo commandinfo)
        {
            m_CCOM.UnregisterCommand(commandinfo);
        }

        public IChatInfo ChatInfoCCOM
        {
            get { return m_ChatInfo_CCOM; }
        }

        public IChatInfo ChatInfoOwner
        {
            get { return m_ChatInfo_Owner; }
        }

        public ICCOMPlugin CCOMPlugin
        {
            get { return m_CCOM; }
        }


        #endregion

        private void buttonOwnerClientProperties_Click(object sender, EventArgs e)
        {
            ((IPlugin)m_Client_Owner).ShowProperties();
        }

        private void buttonCCOMClientProperties_Click(object sender, EventArgs e)
        {
            ((IPlugin)m_Client_CCOM).ShowProperties();
        }

        private void buttonCCOMProperties_Click(object sender, EventArgs e)
        {
            ((IPlugin)m_CCOM).ShowProperties();
        }



        private void listBoxGeneralPlugins_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                Point p = new Point(e.X, e.Y);
                listBoxGeneralPlugins.SelectedIndex = listBoxGeneralPlugins.IndexFromPoint(p);
            }
        }

        private void buttonChangeGeneralPlugins_Click(object sender, EventArgs e)
        {
            (new FormChangeGeneralPlugins(this)).ShowDialog();
        }

        private void FormMain_Load(object sender, EventArgs e)
        {

        }
    }

}