using System;
using System.Collections.Generic;
using System.Text;

namespace LighthouseShared
{
    public delegate bool ChatMessageCallback(IChatInfo chatinfo, ref string screenname, ref string message);
    public delegate bool ChatUserCallback(IChatInfo chatinfo, ref string screenname);

    public interface IPluginHost
    {
        IClientPlugin ClientCCOM { get; }
        IClientPlugin ClientOwner { get; }

        IChatInfo ChatInfoCCOM { get; }
        IChatInfo ChatInfoOwner { get; }

        ICCOMPlugin CCOMPlugin { get; }

        void AddUIPanel(IPlugin plugin, LHPanel panel);
        void RemoveUIPanel(IPlugin plugin, LHPanel panel);

        object RegisterChatMessageHook(IPlugin plugin, ChatMessageCallback callback);
        void UnregisterChatMessageHook(object handle);

        object RegisterChatUserJoinHook(IPlugin plugin, ChatUserCallback callback);
        void UnregisterChatUserJoinHook(object handle);

        object RegisterChatUserLeaveHook(IPlugin plugin, ChatUserCallback callback);
        void UnregisterChatUserLeaveHook(object handle);

        void SendChatMessageByType(string type, string[] items);
        void SendChatMessageByType(string type, string item);
        void SendChatMessageByType(string type, string[] items, bool single);
        void SendChatMessageByType(string type, string item, bool single);

        void SendBan(string screenname);
        void SendUnban(string screenname);
        void SendKick(string screenname);

        ICommandInfo RegisterCommand(string name, string defaulttext, string help, IPlugin plugin, CommandCallback callback);
        void UnregisterCommand(ICommandInfo commandinfo);

        void Log(LogLevel loglevel, string message);
    }
}
