using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;


namespace Oscar5Plugin
{

    class FLAPSocket
    {

        public FLAPSocketType SocketType
        {
            get
            {
                return m_SocketType;
            }
        }
        public Socket RawSocket
        {
            get
            {
                return m_Socket;
            }
        }

        public RateManager RateManager
        {
            get
            {
                return m_RateManager;
            }
        }

        private Socket m_Socket;
        private byte[] m_RecvBuffer;
        private int m_RecvOffset;
        private ushort m_ClientSequence = 0;
        private FLAPSocketType m_SocketType;
        private bool m_Connected = false;
        private RateManager m_RateManager;
        private uint m_LastFNACID = 0;



        public delegate void ReceivePacketHandler(FLAPPacket packet);
        public delegate void DisconnectHandler(FLAPSocket sender);

        public event ReceivePacketHandler OnReceivePacket;
        public event DisconnectHandler OnDisconnect;

        public FLAPSocket(FLAPSocketType type)
        {
            m_SocketType = type;
            m_Socket = new Socket(AddressFamily.InterNetwork, System.Net.Sockets.SocketType.Stream, ProtocolType.Tcp);
            m_RecvBuffer = new byte[6];

            
        }

        public void Connect(string host, int port)
        {
            m_Socket.Connect(host, port);
            m_Connected = true;
            m_RateManager = new RateManager(this);
            Socket_BeginReceive();
        }

        void ProcessDisconnect()
        {
            if (m_Connected)
            {
                m_Connected = false;
                if (OnDisconnect != null)
                    OnDisconnect(this);
            }
        }

        public void Close()
        {
            m_Socket.Close();
            ProcessDisconnect();
        }

        public void QueuePacket(FLAPPacket packet)
        {
            m_RateManager.SendPacket(packet);
        }

        public void SendPacket(FLAPPacket packet)
        {
            byte[] header = new byte[6];
            header[0] = 0x2a;
            header[1] = (byte)packet.Channel;
            Tools.PutWord(header, 2, m_ClientSequence);
            Tools.PutWord(header, 4, (ushort)packet.Data.Length);

            if (m_ClientSequence >= 65535)
                m_ClientSequence = 0;
            else
                m_ClientSequence++;

            
            byte[] tmppacket = new byte[header.Length + packet.Data.Length];

            Buffer.BlockCopy(header, 0, tmppacket, 0, header.Length);
            Buffer.BlockCopy(packet.Data, 0, tmppacket, header.Length, packet.Data.Length);

            Socket_Send(tmppacket);
        }

        private void Socket_Send(byte[] buffer)
        {
            try
            {
                int offset = 0;

                while (offset < buffer.Length)
                {
                    offset += m_Socket.Send(buffer, offset, buffer.Length - offset, SocketFlags.None);
                }
            }
            catch (Exception)
            {
                try
                {
                    m_Socket.Close();
                }
                finally
                {
                    ProcessDisconnect();
                }
            }
        }

        public uint GetNextFNACID()
        {
            if (m_LastFNACID >= UInt32.MaxValue)
                m_LastFNACID = 0;
            else
                m_LastFNACID++; //what the hell does FNAC even mean any way fucking weird ass aol faggots

            return m_LastFNACID;
        }

        private void Socket_BeginReceive()
        {
            try
            {
                m_Socket.BeginReceive(m_RecvBuffer, m_RecvOffset, m_RecvBuffer.Length - m_RecvOffset, SocketFlags.None, new AsyncCallback(Socket_EndReceive), null);
            }
            catch (Exception)
            {
                try
                {
                    m_Socket.Close();
                }
                finally
                {
                    ProcessDisconnect();
                }
            }
        }

        private void Socket_EndReceive(IAsyncResult result)
        {
            try
            {
                int received = m_Socket.EndReceive(result);

                m_RecvOffset += received;

                if (m_RecvOffset >= 6)
                {
                    FLAPPacket packet = new FLAPPacket();
                    packet.Socket = this;
                    packet.Channel = (FLAPChannel)m_RecvBuffer[1];
                    //packet.Sequence = Tools.GetWord(m_RecvBuffer, 2); //Comment this out for efficiency, it isn't really needed since we use TCP
                    packet.Data = new byte[Tools.GetWord(m_RecvBuffer, 4)];

                    int offset = 0;

                    while (offset < packet.Data.Length) // (>^.^>)
                    {
                        offset += m_Socket.Receive(packet.Data, offset, packet.Data.Length - offset, SocketFlags.None);
                    }

                    if (OnReceivePacket != null)
                        OnReceivePacket(packet);

                    m_RecvOffset = 0;
                }

                Socket_BeginReceive();
            }
            catch (Exception)
            {
                try
                {
                    m_Socket.Close();
                }
                finally
                {
                    ProcessDisconnect();
                }
            }
        }
    }

    class FLAPPacket
    {
        public FLAPChannel Channel = FLAPChannel.SNAC;
        public FLAPSocket Socket;
        public ushort Sequence = 0;
        public byte[] Data;
        public int Offset;
    }

    enum FLAPChannel
    {
        Connect = 0x01,
        SNAC = 0x02,
        Error = 0x03,
        Disconnect = 0x04,
        KeepAlive = 0x05
    }

    enum FLAPSocketType
    {
        Auth,
        BOS,
        ChatNavigation,
        Chat
    }

}
