using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Threading;

using LighthouseShared;

namespace Oscar5Plugin
{
    public class Oscar5 : IClientPlugin, IPlugin
    {
        public string Name
        {
            get
            {
                return "Oscar5 Client";
            }
        }

        public Version Version
        {
            get
            {
                return new Version("0.1");
            }
        }

        public string Description
        {
            get
            {
                return "An AIM/OSCAR client by turbzy, using the protocol as captured from AIM 5.5";
            }
        }

        public bool SignedOn
        {
            get
            {
                return m_SignedOn;
            }
        }

        private FLAPSocket m_AuthSocket;
        private FLAPSocket m_BOSSocket;
        private FLAPSocket m_ChatNavSocket;

        private Queue<ChatInfo> m_ChatJoinQueue;
        private Dictionary<uint, ChatInfo> m_ChatNavCreateRequests;
        private Dictionary<uint, ChatInfo> m_ChatNavRedirectRequests;
        private Dictionary<FLAPSocket, ChatInfo> m_ChatSockets;

        private string m_ScreenName;
        public string ScreenName
        {
            get
            {
                return m_ScreenName;
            }
        }
        
        private string m_FormattedScreenName;
        private string m_Password;
        private int m_BOSPort;

        private ulong m_ICBMCookie;
        
        internal string m_Host = "login.oscar.aol.com";
        internal int m_Port = 5190;

        internal bool m_ReceiveChatPackets = true;
        internal bool m_ReceiveIMPackets = true;

        private bool m_SignedOn = false;

        private Dictionary<string, byte[]> m_Cookies;

        #region IClientPlugin Members

        public void SignOn(string screenname, string password)
        {
            m_Cookies = new Dictionary<string, byte[]>();
            m_ChatNavCreateRequests = new Dictionary<uint, ChatInfo>();
            m_ChatJoinQueue = new Queue<ChatInfo>();
            m_ChatNavRedirectRequests = new Dictionary<uint, ChatInfo>();
            m_ChatSockets = new Dictionary<FLAPSocket, ChatInfo>();

            m_ScreenName = screenname.ToLower().Replace(" ", "");
            m_Password = password;

            if (m_AuthSocket != null)
                m_AuthSocket.Close();

            if (OnStatusUpdate != null)
                OnStatusUpdate(this, "Connecting");

            m_AuthSocket = new FLAPSocket(FLAPSocketType.Auth);
            m_AuthSocket.OnDisconnect += new FLAPSocket.DisconnectHandler(Socket_OnDisconnect);
            m_AuthSocket.OnReceivePacket += new FLAPSocket.ReceivePacketHandler(Socket_OnReceivePacket);
            m_AuthSocket.Connect(m_Host, m_Port);


        }

        public void SignOff()
        {
            if (m_AuthSocket != null)
            {
                m_AuthSocket.Close();
                m_AuthSocket = null;
            }

            if (m_BOSSocket != null)
            {
                m_BOSSocket.Close();
                m_BOSSocket = null;
            }

            if (m_ChatNavSocket != null)
            {
                m_ChatNavSocket.Close();
                m_ChatNavSocket = null;
            }

            if (m_ChatSockets != null)
            {
                try
                {
                    foreach (KeyValuePair<FLAPSocket, ChatInfo> chatsocket in m_ChatSockets)
                    {
                        chatsocket.Key.Close();
                        chatsocket.Value.m_PendingBans.Clear();
                        chatsocket.Value.m_PendingUnbans.Clear();
                    }
                }
                catch (Exception)
                {
                }
                finally
                {

                    m_ChatSockets.Clear();
                    m_ChatSockets = null;
                }
            }

            if (m_ChatJoinQueue != null)

            {
                m_ChatJoinQueue.Clear();
                m_ChatJoinQueue = null;
            }

            if (m_ChatNavCreateRequests != null)
            {
                m_ChatNavCreateRequests.Clear();
                m_ChatNavCreateRequests = null;
            }

            if (m_ChatNavRedirectRequests != null)
            {
                m_ChatNavRedirectRequests.Clear();
                m_ChatNavRedirectRequests = null;
            }

            ProcessSignOff();
        }

        void ProcessSignOff()
        {
            if (m_SignedOn)
            {
                m_SignedOn = false;
            }

            if (OnSignedOff != null)
                OnSignedOff(this);
        }

        public void ChatJoin(string roomname, ushort exchange)
        {
            try
            {
                ChatInfo chatinfo = new ChatInfo(this, roomname, exchange);

                m_ChatJoinQueue.Enqueue(chatinfo);

                if (m_ChatNavSocket != null && m_ChatNavSocket.RawSocket.Connected)
                {
                    ProcessChatJoinQueue();
                }
                else
                {
                    m_BOSSocket.QueuePacket(PacketBuilder.SNACGeneric.ServiceRequest(0x000d));
                }
            }
            catch (Exception)
            {
                throw new ClientNotSignedOnException();
            }
        }

        public void ChatLeave(IChatInfo chatinfo)
        {
            try
            {
                FLAPSocket socket = ((ChatInfo)chatinfo).Socket;

                socket.Close();
            }
            catch (Exception)
            {
                if (OnChatLeave != null)
                {
                    OnChatLeave(this, chatinfo, false, false);
                }
            }
        }

        public ulong ChatMessage(IChatInfo chatinfo, string message)
        {
            FLAPSocket socket = ((ChatInfo)chatinfo).Socket;
            ulong icbmcookie = GetNextICBMCookie();

            socket.QueuePacket(PacketBuilder.SNACChat.Message(message, icbmcookie));

            return icbmcookie;
        }

        public void ChatBan(IChatInfo chatinfo, string screenname)
        {
            ChatInfo realchatinfo = (ChatInfo)chatinfo;

            FLAPSocket socket = realchatinfo.Socket;

            BanInfo ban = new BanInfo();
            ban.IsKick = false;
            ban.ScreenName = screenname;
            ban.ChatInfo = realchatinfo;

            uint fnacid = socket.GetNextFNACID();
            
            lock (realchatinfo.m_PendingBans)
            {
                realchatinfo.m_PendingBans.Add(fnacid, ban);
            }

            socket.QueuePacket(PacketBuilder.SNACChat.Ban(screenname, fnacid));
        }

        public void ChatKick(IChatInfo chatinfo, string screenname)
        {
            ChatInfo realchatinfo = (ChatInfo)chatinfo;
            FLAPSocket socket = realchatinfo.Socket;

            BanInfo ban = new BanInfo();
            ban.IsKick = true;
            ban.ScreenName = screenname;
            ban.ChatInfo = realchatinfo;

            uint fnacid = socket.GetNextFNACID();

            lock (realchatinfo.m_PendingBans)
            {
                realchatinfo.m_PendingBans.Add(fnacid, ban);
            }

            socket.QueuePacket(PacketBuilder.SNACChat.Ban(screenname, fnacid));
        }
        


        public void ChatUnban(IChatInfo chatinfo, string screenname)
        {
            ChatInfo realchatinfo = (ChatInfo)chatinfo;
            FLAPSocket socket = realchatinfo.Socket;

            BanInfo unban = new BanInfo();
            unban.IsKick = false;
            unban.ScreenName = screenname;
            unban.ChatInfo = realchatinfo;

            uint fnacid = socket.GetNextFNACID();

            lock (realchatinfo.m_PendingUnbans)
            {
                realchatinfo.m_PendingUnbans.Add(fnacid, unban);
            }

            socket.QueuePacket(PacketBuilder.SNACChat.Unban(screenname, fnacid));
        }

        public void ChatClose(IChatInfo chatinfo)
        {
            FLAPSocket socket = ((ChatInfo)chatinfo).Socket;

            socket.QueuePacket(PacketBuilder.SNACChat.Close());
        }


        public event ChatJoinHandler OnChatJoin;
        public event ChatLeaveHandler OnChatLeave;
        public event ChatMessageHandler OnChatMessage;
        public event ChatUserHandler OnChatOwnerDetect;
        public event ChatUserHandler OnChatBan;
        public event ChatUserHandler OnChatBanFailed;
        public event ChatUserHandler OnChatUnban;
        public event ChatUserHandler OnChatUnbanFailed;
        public event ChatUserHandler OnChatKick;
        public event ChatUserHandler OnChatKickFailed;
        public event ChatUserHandler OnChatUserJoin;
        public event ChatUserHandler OnChatUserLeave;
        public event SignOnHandler OnSignedOn;
        public event SignOffHandler OnSignedOff;
        public event ErrorCodeHandler OnSignOnError;
        public event StatusUpdateHandler OnStatusUpdate;
        public event RawPacketHandler OnRawPacket;


        #endregion




        void Socket_OnDisconnect(FLAPSocket sender)
        {
            switch (sender.SocketType)
            {
                case FLAPSocketType.Auth:
                    //do nothing i guess?
                    break;
                case FLAPSocketType.BOS:
                    ProcessSignOff();
                    break;
                case FLAPSocketType.ChatNavigation:
                    m_ChatNavSocket = null;
                    break;
                case FLAPSocketType.Chat:
                    ChatInfo chatinfo = m_ChatSockets[sender];
                    
                    if (OnChatLeave != null)
                        OnChatLeave(this, chatinfo, false, false);

                    m_ChatSockets.Remove(sender);

                    break;
            }
        }

        void Socket_OnReceivePacket(FLAPPacket packet)
        {
            switch (packet.Channel)
            {
                case FLAPChannel.Connect:
                    if (m_Cookies.ContainsKey(packet.Socket.RawSocket.RemoteEndPoint.ToString()))
                    {
                        packet.Socket.QueuePacket(PacketBuilder.FLAPCookie(m_Cookies[packet.Socket.RawSocket.RemoteEndPoint.ToString()]));
                        m_Cookies.Remove(packet.Socket.RawSocket.RemoteEndPoint.ToString());
                    }
                    else
                    {
                        packet.Socket.QueuePacket(PacketBuilder.FLAPConnection());
                        packet.Socket.QueuePacket(PacketBuilder.SNACSignOn.SignOn(m_ScreenName));
                    }
                    break;
                case FLAPChannel.SNAC:
                    if (OnRawPacket != null)
                    {
                        if (OnRawPacket(this, packet.Data))
                            HandleSNAC(packet);
                    }
                    else
                    {
                        HandleSNAC(packet);
                    }
                    break;
                case FLAPChannel.Disconnect:
                    packet.Socket.Close();
                    break;
            }
        }


        void HandleSNAC(FLAPPacket packet)
        {
            ushort family = Tools.GetWord(packet);
            ushort subfamily = Tools.GetWord(packet);
            SNACFlags snacflags = (SNACFlags)Tools.GetWord(packet);
            uint fnacid = Tools.GetDword(packet);

            if ((snacflags & SNACFlags.TLVBlockInHeader) == SNACFlags.TLVBlockInHeader)
            {
                ushort tlvblocklength = Tools.GetWord(packet);

                packet.Offset += tlvblocklength;
            }

            switch (family)
            {
                case 0x0001:
                    switch (subfamily)
                    {
                        case 0x0003:
                            HandleServerReady(packet);
                            break;
                        case 0x0005:
                            HandleRedirect(packet);
                            break;
                        case 0x0007:
                            HandleRateInfo(packet);
                            break;
                        case 0x000a:
                            HandleRateChange(packet);
                            break;
                        case 0x0018:
                            HandleCapabilitiesAck(packet);
                            break;
                    }
                    break;
                case 0x000d:
                    switch (subfamily)
                    {
                        case 0x0009:
                            HandleChatInfo(packet);
                            break;

                    }
                    break;

                case 0x000e:
                    switch (subfamily)
                    {
                        case 0x0001:
                            HandleChatError(packet);
                            break;
                        case 0x0003:
                            HandleChatUserJoin(packet);
                            break;
                        case 0x0004:
                            HandleChatUserLeave(packet);
                            break;
                        case 0x0026:
                            HandleChatJoined(packet);
                            break;
                        case 0x0006:
                            HandleChatMessage(packet);
                            break;
                        case 0x0030:
                            HandleChatOwnerDetect(packet);
                            break;
                     
                    }
                    break;
                case 0x0017:
                    switch (subfamily)
                    {
                        case 0x0007:
                            byte[] challenge = new byte[Tools.GetWord(packet.Data, 10)];
                            Buffer.BlockCopy(packet.Data, 12, challenge, 0, challenge.Length);
                            if (OnStatusUpdate != null)
                                OnStatusUpdate(this, "Authorizing");
                            packet.Socket.QueuePacket(PacketBuilder.SNACSignOn.Logon(m_ScreenName, m_Password, challenge));
                            break;
                        case 0x0003:
                            HandleLogonReply(packet);
                            break;
                    }
                    break;
                
            }

        }

        void HandleRateChange(FLAPPacket packet)
        {
            packet.Socket.RateManager.ProcessRateChange(packet);
        }

        void HandleLogonReply(FLAPPacket packet)
        {
            string bostemp = null;
            string bosserver = null;
            int bosport = 0;
            byte[] cookie = null;
            bool haserror = false;
            ushort errorcode = 0;
            string errorstring = null;

            int offset = 10;
            int length = 0;
            int type = 0;
            
            while (offset < packet.Data.Length)
            {
                type = Tools.GetWord(packet.Data, offset);
                offset += 2;
                length = Tools.GetWord(packet.Data, offset);
                offset += 2;

                switch (type)
                {
                    case 0x0001:
                        m_FormattedScreenName = Encoding.ASCII.GetString(packet.Data, offset, length);
                        break;
                    case 0x0005:
                        bostemp = Encoding.ASCII.GetString(packet.Data, offset, length);
                        string[] bossplit = bostemp.Split(":".ToCharArray());
                        bosserver = bossplit[0];
                        bosport = Convert.ToInt32(bossplit[1]);
                        break;
                    case 0x0006:
                        cookie = new byte[length];
                        Buffer.BlockCopy(packet.Data, offset, cookie, 0, length);
                        break;
                    case 0x0008:
                        errorcode = Tools.GetWord(packet.Data, offset);
                        haserror = true;
                        break;
                    case 0x0004:
                        errorstring = Encoding.ASCII.GetString(packet.Data, offset, length);
                        break;
                }

                offset += length;
            }

            if (!haserror)
            {
                if (OnStatusUpdate != null)
                    OnStatusUpdate(this, "Authorized");

                m_Cookies.Add(bostemp, cookie);
                packet.Socket.QueuePacket(PacketBuilder.FLAPClose());
                packet.Socket.Close();
                m_BOSPort = bosport;
                m_BOSSocket = new FLAPSocket(FLAPSocketType.BOS);
                m_BOSSocket.OnReceivePacket += new FLAPSocket.ReceivePacketHandler(Socket_OnReceivePacket);
                m_BOSSocket.OnDisconnect += new FLAPSocket.DisconnectHandler(Socket_OnDisconnect);
                m_BOSSocket.Connect(bosserver, bosport);
            }
            else
            {
                if (OnStatusUpdate != null)
                {
                    string status = "Unknown Error";

                    switch (errorcode)
                    {
                        case 5:
                            status = "Incorrect Password";
                            break;
                    }

                    OnStatusUpdate(this, status);
                }

                if (OnSignOnError != null)
                    OnSignOnError(this, errorcode, errorstring);

                SignOff();
            }
        }


        void HandleServerReady(FLAPPacket packet)
        {
            packet.Socket.QueuePacket(PacketBuilder.SNACGeneric.Capabilities(packet.Socket.SocketType, m_ReceiveChatPackets, m_ReceiveIMPackets));
        }
        void HandleCapabilitiesAck(FLAPPacket packet)
        {
            packet.Socket.QueuePacket(PacketBuilder.SNACGeneric.RateInfoRequest());
        }
        void HandleRateInfo(FLAPPacket packet)
        {
            packet.Socket.RateManager.ProcessRateParametersReply(packet);

            packet.Socket.QueuePacket(PacketBuilder.SNACGeneric.RateInfoAck(packet.Socket.RateManager.GetRateClasses()));
            packet.Socket.QueuePacket(PacketBuilder.SNACGeneric.ClientReady(packet.Socket.SocketType, m_ReceiveChatPackets, m_ReceiveIMPackets));

            switch (packet.Socket.SocketType)
            {
                case FLAPSocketType.BOS:
                    if (OnStatusUpdate != null)
                        OnStatusUpdate(this, "Signed On");

                    if (OnSignedOn != null)
                        OnSignedOn(this);
                    break;
                case FLAPSocketType.ChatNavigation:
                    ProcessChatJoinQueue();
                    break;
                case FLAPSocketType.Chat:
                    if (OnChatJoin != null)
                        OnChatJoin(this, m_ChatSockets[packet.Socket]);
                    break;
            }


        }
        void HandleRedirect(FLAPPacket packet)
        {
            uint fnacid = Tools.GetDword(packet.Data, 6);

            int offset = 10;
            ushort type = 0;
            ushort length = 0;

            ushort family = 0;
            ushort version = 0;
            string boshost = null;
            byte[] cookie = null;

            if (packet.Data[offset] == 0x00 && packet.Data[offset + 1] == 0x06)
                offset += 2;

            while (offset < packet.Data.Length)
            {
                type = Tools.GetWord(packet.Data, offset);
                offset += 2;
                length = Tools.GetWord(packet.Data, offset);
                offset += 2;

                switch (type)
                {
                    case 0x0001:
                        version = Tools.GetWord(packet.Data, offset);
                        break;
                    case 0x000d:
                        family = Tools.GetWord(packet.Data, offset);
                        break;
                    case 0x0005:
                        boshost = Encoding.ASCII.GetString(packet.Data, offset, length);
                        break;
                    case 0x0006:
                        cookie = new byte[length];
                        Buffer.BlockCopy(packet.Data, offset, cookie, 0, length);
                        break;
                }

                offset += length;
            }

            string bostemp = boshost + ":" + m_BOSPort.ToString();

            m_Cookies.Add(bostemp, cookie);

            switch (family)
            {
                case 0x000d:
                    m_ChatNavSocket = new FLAPSocket(FLAPSocketType.ChatNavigation);
                    m_ChatNavSocket.OnDisconnect += new FLAPSocket.DisconnectHandler(Socket_OnDisconnect);
                    m_ChatNavSocket.OnReceivePacket += new FLAPSocket.ReceivePacketHandler(Socket_OnReceivePacket);
                    m_ChatNavSocket.Connect(boshost, m_BOSPort);
                    break;
                case 0x000e:
                    if (m_ChatNavRedirectRequests.ContainsKey(fnacid))
                    {
                        ChatInfo chatinfo = m_ChatNavRedirectRequests[fnacid];
                        m_ChatNavRedirectRequests.Remove(fnacid);
                        
                        FLAPSocket chatsocket = new FLAPSocket(FLAPSocketType.Chat);
                        chatsocket.OnReceivePacket += new FLAPSocket.ReceivePacketHandler(Socket_OnReceivePacket);
                        chatsocket.OnDisconnect += new FLAPSocket.DisconnectHandler(Socket_OnDisconnect);
                        m_ChatSockets.Add(chatsocket, chatinfo);
                        chatinfo.Socket = chatsocket;

                        chatsocket.Connect(boshost, m_BOSPort);

                    }
                    break;
                    
            }
                    
        }

        void HandleChatInfo(FLAPPacket packet)
        {
            int offset = 6;
            uint fnacid = Tools.GetDword(packet.Data, offset);
            offset += 4;

            if (m_ChatNavCreateRequests.ContainsKey(fnacid))
            {
                ChatInfo chatinfo = m_ChatNavCreateRequests[fnacid];
                m_ChatNavCreateRequests.Remove(fnacid);

                while (offset < packet.Data.Length)
                {
                    int type = Tools.GetWord(packet.Data, offset);
                    offset += 2;
                    int length = Tools.GetWord(packet.Data, offset);
                    offset += 2;

                    switch (type)
                    {
                        case 0x0004:
                            int inneroffset = offset;

                            inneroffset += 2; //skip the exchange, we already know it, duh!
                            byte strlen = packet.Data[inneroffset];
                            inneroffset++;
                            chatinfo.m_FullName = Encoding.ASCII.GetString(packet.Data, inneroffset, strlen);
                            inneroffset += strlen;
                            chatinfo.m_Instance = Tools.GetWord(packet.Data, inneroffset);
                            inneroffset += 2;

                            inneroffset += 3; // skip some crap cuz i dunno what it does lol

                            while (inneroffset < (offset + length))
                            {
                                int innertype = Tools.GetWord(packet.Data, inneroffset);
                                inneroffset += 2;
                                int innerlength = Tools.GetWord(packet.Data, inneroffset);
                                inneroffset += 2;

                                switch (innertype)
                                {
                                    case 0x006a:
                                        chatinfo.m_DisplayName = Encoding.ASCII.GetString(packet.Data, inneroffset, innerlength);
                                        break;
                                    case 0x00d1:
                                        chatinfo.m_MaxMessageLength = Tools.GetWord(packet.Data, inneroffset);
                                        break;
                                    case 0x00d3:
                                        chatinfo.m_DisplayName = Encoding.ASCII.GetString(packet.Data, inneroffset, innerlength);
                                        break;
                                }
                                inneroffset += innerlength;
                            }
                            break;
                    }

                    offset += length;
                }

                fnacid = m_BOSSocket.GetNextFNACID();

                m_ChatNavRedirectRequests.Add(fnacid, chatinfo);

                m_BOSSocket.QueuePacket(PacketBuilder.SNACGeneric.ChatServiceRequest(chatinfo, fnacid));
            }
        }

        void HandleChatJoined(FLAPPacket packet)
        {
            //OnChatJoin(m_ChatSockets[packet.Socket]);
        }

        void HandleChatMessage(FLAPPacket packet)
        {
            ChatInfo chatinfo = m_ChatSockets[packet.Socket];
            string screenname = null;
            string message = null;

            int offset = 10;

            ulong icbmcookie = Tools.GetQword(packet.Data, offset);
            
            offset += 8;

            offset += 2; //skip channel, dont kare

            while (offset < packet.Data.Length)
            {
                ushort type = Tools.GetWord(packet.Data, offset);
                offset += 2;
                ushort length = Tools.GetWord(packet.Data, offset);
                offset += 2;

                switch (type)
                {
                    case 0x0003:
                        byte snlen = packet.Data[offset];
                        screenname = Encoding.ASCII.GetString(packet.Data, offset + 1, snlen);
                        break;
                    case 0x0005:
                        int inneroffset = offset;

                        while (inneroffset < (offset + length))
                        {
                            ushort innertype = Tools.GetWord(packet.Data, inneroffset);
                            inneroffset += 2;
                            ushort innerlength = Tools.GetWord(packet.Data, inneroffset);
                            inneroffset += 2;

                            switch (innertype)
                            {
                                case 0x0001:
                                    message = Encoding.ASCII.GetString(packet.Data, inneroffset, innerlength);
                                    break;
                            }

                            inneroffset += innerlength;
                        }
                        break;
                }

                offset += length;

            }

            if (OnChatMessage != null)
                OnChatMessage(this, chatinfo, screenname, icbmcookie, message); 
         
        }

        void HandleChatOwnerDetect(FLAPPacket packet)
        {
            ChatInfo chatinfo = m_ChatSockets[packet.Socket];

            byte snlen = packet.Data[10];
            string screenname = Encoding.ASCII.GetString(packet.Data, 11, snlen);

            chatinfo.m_Owner = screenname;

            if (screenname.ToLower().Replace(" ", "") == m_ScreenName)
                chatinfo.m_OwnershipAquired = true;
            else
                chatinfo.m_OwnershipAquired = false;

            if (OnChatOwnerDetect != null)
                OnChatOwnerDetect(this, chatinfo, screenname);
        }

        void HandleChatError(FLAPPacket packet)
        {
            ChatInfo chatinfo = m_ChatSockets[packet.Socket];

            uint fnacid = Tools.GetDword(packet.Data, 6);
            ushort error = Tools.GetWord(packet.Data, 10);

            if (chatinfo.m_PendingBans.ContainsKey(fnacid))
            {
                BanInfo ban = chatinfo.m_PendingBans[fnacid];

                if (error == 0x0000)
                {
                    if (ban.IsKick)
                    {
                        uint newfnacid = packet.Socket.GetNextFNACID();

                        lock (chatinfo.m_PendingUnbans)
                        {
                            chatinfo.m_PendingUnbans.Add(newfnacid, ban);
                        }

                        packet.Socket.SendPacket(PacketBuilder.SNACChat.Unban(ban.ScreenName, newfnacid));
                    }
                    else
                    {
                        if (OnChatBan != null)
                            OnChatBan(this, chatinfo, ban.ScreenName);
                    }
                }
                else if (error == 0x000d)
                {
                    if (OnChatBanFailed != null)
                        OnChatBanFailed(this, chatinfo, ban.ScreenName);
                }

                lock (chatinfo.m_PendingBans)
                {
                    chatinfo.m_PendingBans.Remove(fnacid);
                }
            }
            else if (chatinfo.m_PendingUnbans.ContainsKey(fnacid))
            {
                BanInfo unban = chatinfo.m_PendingUnbans[fnacid];

                if (error == 0x0000)
                {
                    if (unban.IsKick && OnChatKick != null)
                    {
                        if (OnChatKick != null)
                            OnChatKick(this, chatinfo, unban.ScreenName);
                    }
                    else
                    {
                        if (OnChatUnban != null)
                            OnChatUnban(this, chatinfo, unban.ScreenName);
                    }
                }
                else if (error == 0x000d)
                {
                    if (unban.IsKick)
                    {
                        if (OnChatKickFailed != null)
                            OnChatKickFailed(this, chatinfo, unban.ScreenName);
                    }
                    else
                    {
                        if (OnChatUnbanFailed != null)
                            OnChatUnbanFailed(this, chatinfo, unban.ScreenName);
                    }
                }

                lock (chatinfo.m_PendingUnbans)
                {
                    chatinfo.m_PendingUnbans.Remove(fnacid);
                }
            }
        }

        void HandleChatUserJoin(FLAPPacket packet)
        {
            int offset = 10;

            while (offset < packet.Data.Length)
            {
                byte snlen = packet.Data[offset];
                offset++;
                string screenname = Encoding.ASCII.GetString(packet.Data, offset, snlen);
                offset += snlen;
                offset += 2;

                ushort tlvcount = Tools.GetWord(packet.Data, offset);
                offset += 2;

                for (int i = 0; i < tlvcount; i++)
                {
                    offset += 2;
                    ushort length = Tools.GetWord(packet.Data, offset);
                    offset += 2;

                    offset += length;
                }
                if (OnChatUserJoin != null)
                    OnChatUserJoin(this, m_ChatSockets[packet.Socket], screenname);
            }
        }

        void HandleChatUserLeave(FLAPPacket packet)
        {
            int offset = 10;

            while (offset < packet.Data.Length)
            {
                byte snlen = packet.Data[offset];
                offset++;
                string screenname = Encoding.ASCII.GetString(packet.Data, offset, snlen);
                offset += snlen;
                offset += 2;

                ushort tlvcount = Tools.GetWord(packet.Data, offset);
                offset += 2;

                for (int i = 0; i < tlvcount; i++)
                {
                    offset += 2;
                    ushort length = Tools.GetWord(packet.Data, offset);
                    offset += 2;

                    offset += length;
                }
                
                if (OnChatUserLeave != null)
                    OnChatUserLeave(this, m_ChatSockets[packet.Socket], screenname);
            }

        }

        

        private void ProcessChatJoinQueue()
        {
            while (m_ChatJoinQueue.Count > 0)
            {
                uint fnacid = m_ChatNavSocket.GetNextFNACID();

                ChatInfo chatinfo = m_ChatJoinQueue.Dequeue();

                m_ChatNavCreateRequests.Add(fnacid, chatinfo);

                m_ChatNavSocket.QueuePacket(PacketBuilder.SNACChatNav.Create(chatinfo, fnacid));
            }
        }


        private ulong GetNextICBMCookie()
        {
            if (m_ICBMCookie >= ulong.MaxValue)
                m_ICBMCookie = 0;
            else
                m_ICBMCookie++;

            return m_ICBMCookie;
        }


        #region IPlugin Members

        public void Load(IPluginHost host, System.IO.FileStream settingsfile)
        {
            
        }

        public void Unload()
        {
            //lol i do nothing! yay for uselessfunction
        }

        #endregion

        #region IPlugin Members


        public void SaveSettings()
        {
        }

        #endregion

        public void ShowProperties()
        {
            (new FormProperties(this)).ShowDialog();
        }
    }
}
