using System;
using System.Collections.Generic;
using System.Text;

namespace Oscar5Plugin
{
    struct Tools
    {
        public static ushort GetWord(IList<byte> buffer, int offset)
        {
            return (ushort)((buffer[offset] * 256) + buffer[offset + 1]);
        }

        public static ushort GetWord(FLAPPacket packet)
        {
            return (ushort)((packet.Data[packet.Offset++] * 256) + packet.Data[packet.Offset++]);
        }

        public static void PutWord(IList<byte> buffer, int offset, ushort value)
        {
            buffer[offset] = (byte)(value / 256);
            buffer[offset + 1] = (byte)(value % 256);
        }

        public static uint GetDword(IList<byte> buffer, int offset)
        {
            return (uint)((GetWord(buffer, offset) * 65536) + (GetWord(buffer, offset + 2)));
        }

        public static uint GetDword(FLAPPacket packet)
        {
            return (uint)((GetWord(packet) * 65536) + GetWord(packet));
        }

        public static void PutDword(IList<byte> buffer, int offset, uint value)
        {
            Tools.PutWord(buffer, offset, (ushort)(value / 65536));
            Tools.PutWord(buffer, offset + 2, (ushort)(value % 65536));
        }

        public static ulong GetQword(IList<byte> buffer, int offset)
        {
            return (ulong)((Tools.GetDword(buffer, offset) * 4294967296) + Tools.GetDword(buffer, offset + 4));
        }

        public static void PutQword(IList<byte> buffer, int offset, ulong value)
        {
            Tools.PutDword(buffer, offset, (uint)(value / 4294967296));
            Tools.PutDword(buffer, offset + 4, (uint)(value % 4294967296));
        }

        // creates a TLV with a string value
        public static byte[] MakeTLV(ushort type, string value)
        {
            //allocate a new byte array to contain the finished tlv
            byte[] tlv = new byte[4 + value.Length];
            //put the type into the array
            Tools.PutWord(tlv, 0, type);
            //put the length into the array
            Tools.PutWord(tlv, 2, (ushort)value.Length);
            //using ISO-8859-1 encoding (what most people call ASCII), put the string value into the array
            System.Text.Encoding.GetEncoding("iso-8859-1").GetBytes(value, 0, value.Length, tlv, 4);

            return tlv;
        }

        public static byte[] MakeTLV(ushort type, byte[] value)
        {
            byte[] tlv = new byte[4 + value.Length];
            Tools.PutWord(tlv, 0, type);
            Tools.PutWord(tlv, 2, (ushort)value.Length);
            Buffer.BlockCopy(value, 0, tlv, 4, value.Length);
            return tlv;
        }

        public static void PutSNACHeader(IList<byte> buffer, ushort family, ushort subfamily, ushort fnacflags, uint fnacid)
        {
            Tools.PutWord(buffer, 0, family);
            Tools.PutWord(buffer, 2, subfamily);
            Tools.PutWord(buffer, 4, fnacflags);
        }




    }
}
