using System;
using System.Collections.Generic;
using System.Text;
using LighthouseShared;

namespace SimpleAutoban
{
    public class SimpleAB : IPlugin, IGeneralPlugin
    {
        public IPluginHost m_PluginHost;
        public System.IO.FileStream m_SettingsFile;

        public Dictionary<string, string> m_Autobans;
        public Dictionary<string, string> m_PartialAutobans;
        public Dictionary<string, string> m_Wordbans;

        public PanelAutoban m_Panel;

        #region IPlugin Members

        public string Name
        {
            get { return "SimpleAutoban"; }
        }

        public Version Version
        {
            get { return new Version("1.0"); }
        }

        public string Description
        {
            get { return "Basic autobanning plugin. Bans ghosters! :)"; }
        }

        public void Load(IPluginHost host, System.IO.FileStream settingsfile)
        {
            m_PluginHost = host;
            m_SettingsFile = settingsfile;

            m_Autobans = new Dictionary<string, string>();
            m_PartialAutobans = new Dictionary<string, string>();
            m_Wordbans = new Dictionary<string, string>();

            m_Panel = new PanelAutoban(this);

            m_PluginHost.AddUIPanel(this, m_Panel);

            m_PluginHost.RegisterChatMessageHook(this, new ChatMessageCallback(ChatMessage_Callback));
            m_PluginHost.RegisterChatUserJoinHook(this, new ChatUserCallback(ChatUserJoined_Callback));

            m_PluginHost.RegisterCommand("AddWordban", "aw", "Add wordban", this, new CommandCallback(DoCCOMSingleArg));
            m_PluginHost.RegisterCommand("RemoveWordban", "unaw", "Remove wordban", this, new CommandCallback(DoCCOMSingleArg));
            m_PluginHost.RegisterCommand("AddAutoban", "ab", "Add autoban", this, new CommandCallback(DoCCOMSplitArgs));
            m_PluginHost.RegisterCommand("RemoveAutoban", "unab", "Remove autoban", this, new CommandCallback(DoCCOMSplitArgs));
        }

        void DoCCOMSingleArg(ICommandInfo commandinfo, IUserInfo userinfo, string args)
        {
            switch (commandinfo.Name)
            {
                case "AddWordban":
                    CCOMAddWordban(args);
                    break;
                case "RemoveWordban":
                    CCOMRemoveWordban(args);
                    break;
            }
        }

        void DoCCOMSplitArgs(ICommandInfo commandinfo, IUserInfo userinfo, string args)
        {
            string[] argsplit = args.ToLower().Replace(" ", "").Split(",".ToCharArray());

            foreach (string arg in argsplit)
            {
                switch (commandinfo.Name)
                {
                    case "AddAutoban":
                        CCOMAddAutoban(arg);
                        break;
                    case "RemoveAutoban":
                        CCOMRemoveAutoban(arg);
                        break;
                }
            }
        }


        public bool ChatMessage_Callback(IChatInfo chatinfo, ref string screenname, ref string message)
        {
            screenname = screenname.ToLower().Replace(" ", "");

            IUserInfo userinfo = null;
            bool founduser = m_PluginHost.CCOMPlugin.List_Users.TryGetValue(screenname, out userinfo);

            if (m_Panel.checkBanGhost.Checked)
            {
                if (!founduser)
                {
                    m_PluginHost.SendBan(screenname);
                }
                else
                {
                    if (!userinfo.Here && !userinfo.Immune)
                    {
                        m_PluginHost.SendBan(screenname);
                    }
                }
            }
            else if (m_Panel.checkEnableWordban.Checked)
            {
                message = System.Text.RegularExpressions.Regex.Replace(message, @"<(.|\n)*?>", string.Empty);
                if (!founduser || !userinfo.Immune)
                {
                    foreach (string wordban in m_Wordbans.Values)
                    {
                        if (message.Contains(wordban))
                        {
                            m_PluginHost.SendBan(screenname);
                            break;
                        }
                    }
                }
            }

            return false;
        }

        public bool ChatUserJoined_Callback(IChatInfo chatinfo, ref string screenname)
        {
            screenname = screenname.ToLower().Replace(" ", "");

            IUserInfo userinfo = m_PluginHost.CCOMPlugin.List_Users[screenname];

            if (!userinfo.Immune)
            {
                if (m_Panel.radioBanAll.Checked)
                {
                    m_PluginHost.SendBan(screenname);
                    return false;
                }
                else if (m_Panel.radioBanList.Checked)
                {
                    if (m_Autobans.ContainsKey(screenname))
                    {
                        m_PluginHost.SendBan(screenname);
                        return false;
                    }
                    else
                    {
                        foreach (string s in m_PartialAutobans.Values)
                        {
                            if (screenname.Contains(s))
                            {
                                m_PluginHost.SendBan(screenname);
                                return false;
                            }
                        }
                    }
                }

                if (m_Panel.checkBanICQ.Checked)
                {
                    uint sntouint;

                    if (uint.TryParse(screenname, out sntouint))
                    {
                        m_PluginHost.SendBan(screenname);
                    }
                }
            }

            return false;
        }

        public void CCOMAddAutoban(string autoban)
        {
            autoban = autoban.ToLower().Replace(" ", "");

            if (autoban.Contains("*"))
            {
                autoban = autoban.Replace("*", "");

                if (autoban.Length > 0 || m_Panel.checkAllowABStar.Checked)
                {
                    if (!m_PartialAutobans.ContainsKey(autoban))
                    {
                        m_PartialAutobans.Add(autoban, autoban);
                        m_PluginHost.SendChatMessageByType("Add Partial Autoban", "*" + autoban);
                    }
                    else
                    {
                        m_PluginHost.SendChatMessageByType("Error", "Cannot add partial autoban  already in list", true);
                    }
                }
                else
                {
                    m_PluginHost.SendChatMessageByType("Error", "Cannot add partial autoban of 0 length");
                }
            }
            else
            {
                if (autoban.Length > 0)
                {
                    if (!m_Autobans.ContainsKey(autoban))
                    {
                        m_Autobans.Add(autoban, autoban);
                        m_PluginHost.SendChatMessageByType("Add Autoban", autoban);
                    }
                    else
                    {
                        m_PluginHost.SendChatMessageByType("Error", "Cannot add autoban  already in list", true);
                    }
                }
                else
                {
                    m_PluginHost.SendChatMessageByType("Error", "Cannot add autoban of 0 length");
                }
            }
        }

        public void CCOMRemoveAutoban(string autoban)
        {
            autoban = autoban.ToLower().Replace(" ", "");

            if (autoban.Contains("*"))
            {
                autoban = autoban.Replace("*", "");

                if (autoban.Length > 0 || m_Panel.checkAllowABStar.Checked)
                {
                    if (m_PartialAutobans.ContainsKey(autoban))
                    {
                        m_PartialAutobans.Remove(autoban);
                        m_PluginHost.SendChatMessageByType("Remove Partial Autoban", "*" + autoban);
                    }
                    else
                    {
                        m_PluginHost.SendChatMessageByType("Error", "Cannot remove partial autoban  not in list", true);
                    }
                }
                else
                {
                    m_PluginHost.SendChatMessageByType("Error", "Cannot remove partial autoban of 0 length");
                }
            }
            else
            {
                if (autoban.Length > 0)
                {
                    if (m_Autobans.ContainsKey(autoban))
                    {
                        m_Autobans.Remove(autoban);
                        m_PluginHost.SendChatMessageByType("Remove Autoban", autoban);
                    }
                    else
                    {
                        m_PluginHost.SendChatMessageByType("Error", "Cannot remove partial autoban  not in list", true);
                    }
                }
                else
                {
                    m_PluginHost.SendChatMessageByType("Error", "Cannot remove autoban of 0 length");
                }
            }
        }

        public void CCOMAddWordban(string wordban)
        {
            wordban = wordban.ToLower().Replace(" ", "");

            if (wordban.Length > 0)
            {
                if (!m_Wordbans.ContainsKey(wordban))
                {
                    m_Wordbans.Add(wordban, wordban);
                    m_PluginHost.SendChatMessageByType("Add Wordban", wordban);
                }
                else
                {
                    m_PluginHost.SendChatMessageByType("Error", "Cannot add wordban  already in list");
                }
            }
            else
            {
                m_PluginHost.SendChatMessageByType("Error", "Cannot add wordban of 0 length");
            }
        }

        public void CCOMRemoveWordban(string wordban)
        {
            wordban = wordban.ToLower().Replace(" ", "");

            if (wordban.Length > 0)
            {
                if (m_Wordbans.ContainsKey(wordban))
                {
                    m_Wordbans.Remove(wordban);
                    m_PluginHost.SendChatMessageByType("Remove Wordban", wordban);
                }
                else
                {
                    m_PluginHost.SendChatMessageByType("Error", "Cannot remove wordban  not in list");
                }
            }
            else
            {
                m_PluginHost.SendChatMessageByType("Error", "Cannot remove wordban of 0 length");
            }
        }

        public void Unload()
        {
            m_SettingsFile.Close();
        }

        public void ShowProperties()
        {

        }

        public void SaveSettings()
        {
            
        }

        #endregion
    }
}
